/*
 * DoubleCheck, a UCI chess playing engine. Copyright (C) 2011 Lucas Braesch.
 * 
 * DoubleCheck is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * DoubleCheck is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program. If not,
 * see <http://www.gnu.org/licenses/>.
 * 
 * Credits: Magic bitboard code from Pradu Kannan.
*/
#pragma once
#include "defs.h"

void init_bitboard();

extern bool BitboardInitialized;

extern U64
	Bit[NB_SQUARE],
    Rank[NB_RANK_FILE], File[NB_RANK_FILE],
    Between[NB_SQUARE][NB_SQUARE], Direction[NB_SQUARE][NB_SQUARE],
    KAttacks[NB_SQUARE], NAttacks[NB_SQUARE], PAttacks[NB_COLOR][NB_SQUARE],
    PInitialRank[NB_COLOR], PPromotionRank[NB_COLOR];

extern U64
	AdjacentFiles[NB_RANK_FILE],
	ThisAndAdjacentFiles[NB_RANK_FILE],
	InFront[NB_COLOR][NB_RANK_FILE],
	SquaresInFront[NB_COLOR][NB_SQUARE],
	PassedPawnMask[NB_COLOR][NB_SQUARE];

static inline U64 file_bb(int sq) { return File[file(sq)]; }
static inline U64 rank_bb(int sq) { return Rank[rank(sq)]; }
static inline U64 adjacent_files_bb(int sq) { return AdjacentFiles[file(sq)]; }
static inline U64 this_and_adjacent_files_bb(int sq) { return ThisAndAdjacentFiles[file(sq)]; }
static inline U64 in_front_bb(int color, int sq) { assert(color_ok(color)); return InFront[color][rank(sq)]; }

void print_bitboard(U64 b);

static inline void set_bit(U64 *b, int i) { assert(square_ok(i)); *b |= Bit[i]; }
static inline void clear_bit(U64 *b, int i) { assert(square_ok(i)); *b &= ~Bit[i]; }
static inline U64 test_bit(U64 b, int i) { assert(square_ok(i)); return b & Bit[i]; }
static inline U64 shift_bit(U64 b, int i) { return i > 0 ? b << i : b >> -i; }

extern const int BitTable[NB_SQUARE];

static inline int first_bit(U64 b)
{
	return BitTable[((b & -b) * 0x218a392cd3d5dbfULL) >> 58];
}

static inline int next_bit(U64 *b)
{
	U64 _b = *b;
	*b &= *b - 1;
	return BitTable[((_b & -_b) * 0x218a392cd3d5dbfULL) >> 58];
}

static inline int count_bit(U64 b)
{
	b -= ((b>>1) & 0x5555555555555555ULL);
	b = ((b>>2) & 0x3333333333333333ULL) + (b & 0x3333333333333333ULL);
	b = ((b>>4) + b) & 0x0F0F0F0F0F0F0F0FULL;
	b *= 0x0101010101010101ULL;
	return b >> 56;
}

static inline int count_bit_max15(U64 b)
{
	b -= (b>>1) & 0x5555555555555555ULL;
	b = ((b>>2) & 0x3333333333333333ULL) + (b & 0x3333333333333333ULL);
	b *= 0x1111111111111111ULL;
	return b >> 60;
}

/* Magic bitboards */

extern U64 magic_bishop_db[64][1 << 9];
extern const U64 magicmoves_b_magics[64], magicmoves_b_mask[64];

extern U64 magic_rook_db[64][1 << 12];
extern const U64 magicmoves_r_magics[64], magicmoves_r_mask[64];

extern U64 BPseudoAttacks[NB_SQUARE], RPseudoAttacks[NB_SQUARE];

static inline U64 bishop_attack(int square, U64 occ)
{ return magic_bishop_db[square][((occ & magicmoves_b_mask[square]) * magicmoves_b_magics[square]) >> 55]; }

static inline U64 rook_attack(int square, U64 occ)
{ return magic_rook_db[square][((occ & magicmoves_r_mask[square]) * magicmoves_r_magics[square]) >> 52]; }

static inline U64 piece_attack(int piece, int sq, U64 occ)
{
	assert(Knight <= piece && piece <= King && square_ok(sq));

	if (piece == Knight)
		return NAttacks[sq];
	else if (piece == Bishop)
		return bishop_attack(sq, occ);
	else if (piece == Rook)
		return rook_attack(sq, occ);
	else if (piece == Queen)
		return bishop_attack(sq, occ) | rook_attack(sq, occ);
	else	// if (piece == King)
		return KAttacks[sq];
}